import logging
import warnings
from collections import OrderedDict

import torch
import torch.nn as nn

from ..constants import ALPHABET_UNMOD
from ..data.processing.feature_extractors import FEATURE_EXTRACTORS_PARAMETERS
from ..layers.attention_torch import AttentionLayer
from ..layers.bi_gru_seq_encoder_torch import BiGRUSequentialEncoder
from ..layers.gru_seq_decoder_torch import GRUSequentialDecoder

logger = logging.getLogger("dlomix.models.prosit_torch")


class PrositRetentionTimePredictor(nn.Module):
    """
    Implementation of the Prosit model for retention time prediction.

    Parameters
    -----------
    embedding_output_dim : int, optional
        Size of the embeddings to use. Defaults to 16.
    seq_length : int, optional
        Sequence length of the peptide sequences. Defaults to 30.
    alphabet : dict, optional
        Dictionary mapping for the alphabet (the amino acids in this case). Defaults to ALPHABET_UNMOD.
    dropout_rate : float, optional
        Probability to use for dropout layers in the encoder. Defaults to 0.5.
    latent_dropout_rate : float, optional
        Probability to use for dropout layers in the regressor layers after encoding. Defaults to 0.1.
    recurrent_layers_sizes : tuple, optional
        A tuple of 2 values for the sizes of the two GRU layers in the encoder. Defaults to (256, 512).
    regressor_layer_size : int, optional
        Size of the dense layer in the regressor after the encoder. Defaults to 512.


    """

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        alphabet=ALPHABET_UNMOD,
        dropout_rate=0.5,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
    ):
        super(PrositRetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(alphabet)

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes
        self.embedding_output_dim = embedding_output_dim
        self.seq_length = seq_length

        self.embedding = nn.Embedding(
            num_embeddings=self.embeddings_count,
            embedding_dim=embedding_output_dim,
            padding_idx=0,  # TODO check this
        )

        self.encoder = BiGRUSequentialEncoder(
            embedding_output_dim, self.recurrent_layers_sizes, self.dropout_rate
        )

        self.attention = AttentionLayer(
            feature_dim=self.recurrent_layers_sizes[1], seq_len=self.seq_length
        )

        self.regressor = nn.Sequential(
            OrderedDict(
                [
                    (
                        "dense",
                        nn.Linear(
                            in_features=self.recurrent_layers_sizes[1],
                            out_features=self.regressor_layer_size,
                        ),
                    ),
                    ("activation_relu", nn.ReLU()),
                    ("regressor_dropout", nn.Dropout(self.latent_dropout_rate)),
                ]
            )
        )

        self.output_layer = nn.Linear(
            in_features=self.regressor_layer_size, out_features=1
        )

    def forward(self, inputs, **kwargs):
        x = self.embedding(inputs)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        return x


class PrositIntensityPredictor(nn.Module):
    """
    Implementation of the Prosit model for fragment ion intensity prediction.

    Parameters
    -----------





    """

    DEFAULT_INPUT_KEYS = {
        "SEQUENCE_KEY": "sequence",
        "COLLISION_ENERGY_KEY": "collision_energy",
        "PRECURSOR_CHARGE_KEY": "precursor_charge",
        "FRAGMENTATION_TYPE_KEY": "fragmentation_type",
    }

    # can be extended to include all possible meta data
    META_DATA_KEYS = [
        "COLLISION_ENERGY_KEY",
        "PRECURSOR_CHARGE_KEY",
        "FRAGMENTATION_TYPE_KEY",
    ]

    # retrieve the Lookup PTM feature keys
    PTM_INPUT_KEYS = [*FEATURE_EXTRACTORS_PARAMETERS.keys()]

    def __init__(
        self,
        embedding_output_dim=16,
        seq_length=30,
        len_fion=6,
        alphabet=None,
        dropout_rate=0.2,
        latent_dropout_rate=0.1,
        recurrent_layers_sizes=(256, 512),
        regressor_layer_size=512,
        use_prosit_ptm_features=False,
        input_keys=None,
        meta_data_keys=None,
        with_termini=True,
    ):
        super(PrositIntensityPredictor, self).__init__()

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes
        self.embedding_output_dim = embedding_output_dim
        self.seq_length = seq_length
        self.len_fion = len_fion
        self.use_prosit_ptm_features = use_prosit_ptm_features
        self.input_keys = input_keys
        self.meta_data_keys = meta_data_keys

        # maximum number of fragment ions
        self.max_ion = self.seq_length - 1

        # account for encoded termini
        if with_termini:
            self.max_ion = self.max_ion - 2

        if alphabet:
            self.alphabet = alphabet
        else:
            self.alphabet = ALPHABET_UNMOD

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(self.alphabet) + 1

        self.embedding = nn.Embedding(
            num_embeddings=self.embeddings_count,
            embedding_dim=self.embedding_output_dim,
        )

        self._build_encoders()
        self._build_decoder()

        self.attention = AttentionLayer(
            feature_dim=regressor_layer_size, seq_len=seq_length
        )

        self.meta_data_fusion_layer = None
        if self.meta_data_keys:
            self.meta_data_fusion_layer = MetaDataFusionBlock(max_ion=self.max_ion)

        self.regressor = nn.Sequential(
            OrderedDict(
                [
                    ("time_dense", nn.LazyLinear(out_features=len_fion)),
                    ("activation", nn.LeakyReLU()),
                    ("output", nn.Flatten()),
                ]
            )
        )

    def _build_encoders(self):
        # sequence encoder -> always present
        gru_features_input_size = self.embedding_output_dim

        if self.use_prosit_ptm_features:
            gru_features_input_size = self.embedding_output_dim * 2

        self.sequence_encoder = BiGRUSequentialEncoder(
            embedding_output_dim=gru_features_input_size,
            recurrent_layers_sizes=self.recurrent_layers_sizes,
            dropout_rate=self.dropout_rate,
        )

        # meta data encoder -> optional, only if meta data keys are provided
        self.meta_encoder = None
        if self.meta_data_keys:
            self.meta_encoder = nn.Sequential(
                OrderedDict(
                    [
                        (
                            "meta_dense",
                            nn.LazyLinear(out_features=self.recurrent_layers_sizes[1]),
                        ),
                        ("dropout", nn.Dropout(p=self.dropout_rate)),
                    ]
                )
            )

        # ptm encoder -> optional, only if ptm flag is provided
        self.ptm_input_encoder, self.ptm_aa_fusion = None, None
        if self.use_prosit_ptm_features:
            self.ptm_input_encoder = nn.Sequential(
                Concatenate(dim=-1),
                nn.LazyLinear(out_features=self.regressor_layer_size // 2),
                nn.Dropout(p=self.dropout_rate),
                nn.LazyLinear(out_features=self.embedding_output_dim * 4),
                nn.Dropout(p=self.dropout_rate),
                nn.LazyLinear(out_features=self.embedding_output_dim),
                nn.Dropout(p=self.dropout_rate),
            )

            self.ptm_aa_fusion = Concatenate(dim=-1)

    def _build_decoder(self):
        self.decoder = GRUSequentialDecoder(
            recurrent_layers_sizes=self.recurrent_layers_sizes,
            dropout_rate=self.dropout_rate,
            max_ion=self.max_ion,
        )

    def forward(self, inputs, **kwargs):
        encoded_meta = None
        encoded_ptm = None

        if not isinstance(inputs, dict):
            # when inputs has (seq, target), it comes as tuple
            peptides_in = inputs
        else:
            peptides_in = inputs.get(self.input_keys["SEQUENCE_KEY"])

            # read meta data from the input dict
            # note that the value here is the key to use in the inputs dict passed from the dataset
            meta_data = self._collect_values_from_inputs_if_exists(
                inputs, self.meta_data_keys
            )

            if self.meta_encoder and len(meta_data) > 0:
                if isinstance(meta_data, list):
                    meta_data = torch.cat(meta_data, dim=-1)
                encoded_meta = self.meta_encoder(meta_data)

            elif self.meta_data_keys:
                raise ValueError(
                    f"Following metadata keys were specified when creating the model: {self.meta_data_keys}, but the corresponding values do not exist in the input. The actual input passed to the model contains the following keys: {list(inputs.keys())}"
                )
            else:
                pass

            # ToDo: ensure PTM features work as expected
            # read PTM features from the input dict # --> Still needs to be implemented
            ptm_ac_features = self._collect_values_from_inputs_if_exists(
                inputs, PrositIntensityPredictor.PTM_INPUT_KEYS
            )

            if self.ptm_input_encoder and len(ptm_ac_features) > 0:
                logger.debug("PTM features: ")
                for f in ptm_ac_features:
                    logger.debug(f.shape)
                encoded_ptm = self.ptm_input_encoder(ptm_ac_features)
            elif self.use_prosit_ptm_features:
                warnings.warn(
                    f"PTM features enabled and following PTM features are expected in the model for Prosit Intesity: {PrositIntensityPredictor.PTM_INPUT_KEYS}. The actual input passed to the model contains the following keys: {list(inputs.keys())}. Falling back to no PTM features."
                )

        x = self.embedding(peptides_in)

        # fusion of PTMs (before going into the GRU sequence encoder)
        if self.ptm_aa_fusion and encoded_ptm is not None:
            logger.debug("before ptm fusion: %s", x.shape)
            logger.debug("before ptm fusion enc ptm: %s", encoded_ptm.shape)
            x = self.ptm_aa_fusion([x, encoded_ptm])
            logger.debug("concatednated after fusion: %s", x.shape)

        x = self.sequence_encoder(x)

        x = self.attention(x)

        if self.meta_data_fusion_layer and encoded_meta is not None:
            x = self.meta_data_fusion_layer([x, encoded_meta])
        else:
            # no metadata -> add a dimension to comply with the shape
            x = torch.unsqueeze(x, axis=1)

        x = self.decoder(x)

        x = self.regressor(x)

        return x

    def _collect_values_from_inputs_if_exists(self, inputs, keys_mapping):
        collected_values = []

        keys = []
        if isinstance(keys_mapping, dict):
            keys = keys_mapping.values()

        elif isinstance(keys_mapping, list):
            keys = keys_mapping

        for key_in_inputs in keys:
            # get the input under the specified key if exists
            single_input = inputs.get(key_in_inputs, None)
            if single_input is not None:
                if single_input.ndim == 1:
                    single_input = torch.unsqueeze(single_input, axis=-1)
                collected_values.append(single_input)
        return collected_values


class Concatenate(nn.Module):
    def __init__(self, dim=-1):
        super(Concatenate, self).__init__()
        self.dim = dim

    def forward(self, inputs):
        for x in inputs:
            logger.debug("concatenating: %s ", x.shape)
        return torch.cat(inputs, dim=self.dim)


class MetaDataFusionBlock(torch.nn.Module):
    def __init__(self, max_ion):
        super(MetaDataFusionBlock, self).__init__()
        self.max_ion = max_ion

    def forward(self, x):
        #  x is a tuple of (features, metadata)
        features, metadata = x
        # Multiply operation
        multiplied = features * metadata
        # RepeatVector equivalent - expand along dimension to repeat max_ion times
        repeated = multiplied.unsqueeze(1).expand(-1, self.max_ion, -1)
        return repeated



from os.path import join

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.ticker import LogLocator

from ..data.processing.processors import SequenceParsingProcessor
from .postprocessing import normalize_intensity_predictions
from .Report import PDFFile, Report


class IntensityReport(Report):
    """Report generation for Fragment Ion Intensity Prediction tasks."""

    TARGETS_LABEL = "x"
    PREDICTIONS_LABEL = "y"
    DEFAULT_BATCH_SIZE = 600
    PREDICTIONS_COL_NAME = "intensities_pred"

    def __init__(self, output_path, history, figures_ext="png", batch_size=0):
        super(IntensityReport, self).__init__(output_path, history, figures_ext)

        self.pdf_file = PDFFile("DLOmix - Fragment Ion Intensity Report")

        if batch_size:
            self.batch_size = batch_size
        else:
            self.batch_size = IntensityReport.DEFAULT_BATCH_SIZE

    def generate_report(
        self,
        dataset,
        predictions,
        split="test",
        precursor_charge_column_name="precursor_charge_onehot",
    ):
        self._init_report_resources()

        self.plot_all_metrics()

        # make custom plots
        self.plot_spectral_angle(
            dataset, predictions, split, precursor_charge_column_name
        )

        self._compile_report_resources_add_pdf_pages()
        self.pdf_file.output(join(self._output_path, "intensity_Report.pdf"), "F")

    def plot_spectral_angle(
        self, dataset, predictions, split, precursor_charge_column_name
    ):
        """Create spectral  plot

        Arguments
        ---------
        dataset:  FragmentIonIntensityDataset
        predictions:  array of predictions
        split:  str for the split name in the FragmentIonIntensityDataset
        precursor_charge_column_name:  str
        """

        predictions_df = (
            dataset[split]
            .select_columns(
                [
                    SequenceParsingProcessor.PARSED_COL_NAMES["seq"],
                    dataset.label_column,
                    *dataset.model_features,
                ]
            )
            .to_pandas()
        )

        predictions_df[IntensityReport.PREDICTIONS_COL_NAME] = predictions.tolist()

        predictions_acc = normalize_intensity_predictions(
            predictions_df,
            sequence_column_name=SequenceParsingProcessor.PARSED_COL_NAMES["seq"],
            labels_column_name=dataset.label_column,
            predictions_column_name=IntensityReport.PREDICTIONS_COL_NAME,
            precursor_charge_column_name=precursor_charge_column_name,
            batch_size=self.batch_size,
        )

        self.prediction_results = predictions_acc

        violin_plot = sns.violinplot(predictions_acc["spectral_angle"])

        save_path = join(
            self._output_path, "violin_spectral_angle_plot" + self._figures_ext
        )

        fig = violin_plot.get_figure()
        fig.savefig(save_path)

        self._add_report_resource(
            "spectral_angle_plot",
            "Spectral angle plot",
            "The following figure shows the spectral angle plot for the test data.",
            save_path,
        )



import logging
import urllib.request
import zipfile
from os import makedirs
from os.path import exists, join

import pytest

from dlomix.data.processing import (
    FunctionProcessor,
    SequenceEncodingProcessor,
    SequencePaddingProcessor,
    SequenceParsingProcessor,
    SequencePTMRemovalProcessor,
)

logger = logging.getLogger(__name__)

SEQ_COLUMN = "sequence"

SEQUENCE_UNMODIFIED = "[]-DEL-[]"
PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI = ["D", "E", "L"]

SEQUENCE_MODIFIED = "[]-HC[UNIMOD:4]VD-[]"
PARSED_SEQUENCE_MODIFIED_NO_TERMINI = ["H", "C[UNIMOD:4]", "V", "D"]

SEQUENCE_MODIFIED_WITH_N_MOD = "[UNIMOD:737]-ILC[UNIMOD:4]SIQGFK[UNIMOD:737]D-[]"
PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI = [
    "I",
    "L",
    "C[UNIMOD:4]",
    "S",
    "I",
    "Q",
    "G",
    "F",
    "K[UNIMOD:737]",
    "D",
]


def assert_parsed_data(
    parsed_data, new_sequence_column_value, parsed_sequence_value, n_term, c_term
):
    assert SequenceParsingProcessor.PARSED_COL_NAMES["seq"] in parsed_data.keys()
    assert SequenceParsingProcessor.PARSED_COL_NAMES["n_term"] in parsed_data.keys()
    assert SequenceParsingProcessor.PARSED_COL_NAMES["c_term"] in parsed_data
    assert SEQ_COLUMN in parsed_data.keys()

    assert (
        parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["seq"]]
        == parsed_sequence_value
    )
    assert parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["n_term"]] == n_term
    assert parsed_data[SequenceParsingProcessor.PARSED_COL_NAMES["c_term"]] == c_term
    assert parsed_data[SEQ_COLUMN] == new_sequence_column_value


def test_sequence_parsing_processor_unmodified():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=False)
    input_data = {SEQ_COLUMN: SEQUENCE_UNMODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_batched():
    p = SequenceParsingProcessor(
        sequence_column_name=SEQ_COLUMN, with_termini=False, batched=True
    )
    input_data = {SEQ_COLUMN: [SEQUENCE_UNMODIFIED]}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        [PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI],
        [PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI],
        ["[]-"],
        ["-[]"],
    )


def test_sequence_parsing_processor_with_termini():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_UNMODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[]-", *PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_with_modifications():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_MODIFIED}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[]-", *PARSED_SEQUENCE_MODIFIED_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_MODIFIED_NO_TERMINI,
        "[]-",
        "-[]",
    )


def test_sequence_parsing_processor_with_modifications_and_nterm_mods():
    p = SequenceParsingProcessor(sequence_column_name=SEQ_COLUMN, with_termini=True)
    input_data = {SEQ_COLUMN: SEQUENCE_MODIFIED_WITH_N_MOD}

    parsed = p(input_data)
    logger.info(parsed)

    assert_parsed_data(
        parsed,
        ["[UNIMOD:737]-", *PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI, "-[]"],
        PARSED_SEQUENCE_MODIFIED_WITH_N_MOD_NO_TERMINI,
        "[UNIMOD:737]-",
        "-[]",
    )


def test_sequence_padding_processor_keep():
    length = 5
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=length)
    input_data = {SEQ_COLUMN: PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI}
    padded = p(input_data)
    logger.info(padded)

    assert len(padded[SEQ_COLUMN]) == length
    assert padded[SEQ_COLUMN] == ["D", "E", "L", 0, 0]
    assert padded[SequencePaddingProcessor.KEEP_COLUMN_NAME]


def test_sequence_padding_processor_drop():
    length = 2
    p = SequencePaddingProcessor(sequence_column_name=SEQ_COLUMN, max_length=length)
    input_data = {SEQ_COLUMN: PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI}
    padded = p(input_data)
    logger.info(padded)

    assert padded[SEQ_COLUMN] == PARSED_SEQUENCE_UNMODIFIED_NO_TERMINI[:length]
    assert not padded[SequencePaddingProcessor.KEEP_COLUMN_NAME]


def test_sequence_encoding_processor():
    pass


def test_sequence_encoding_processor_with_fixed_alphabet():
    pass


def test_sequence_encoding_processor_with_extend_alphabet_enabled():
    pass


def test_sequence_encoding_processor_with_fallback_enabled():
    pass


def test_sequence_ptm_removal_processor():
    pass


def test_function_processor():
    pass



import numpy as np

CLASSES_LABELS = ["Non-Flyer", "Weak Flyer", "Intermediate Flyer", "Strong Flyer"]

alphabet = [
    "0",
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]

aa_to_int_dict = dict((aa, i) for i, aa in enumerate(alphabet))

int_to_aa_dict = dict((i, aa) for i, aa in enumerate(alphabet))

padding_char = np.zeros(len(alphabet))
padding_char[0] = 1
